#include "stdafx.h"
#include "IRPoCom.h"
#include "IRPoUsbData.h"

IRPoCom::IRPoCom()
{
}


IRPoCom::~IRPoCom()
{
  mHid.Close();
}


void IRPoCom::Open()
{
  mHid.Close();
  if (!mHid.Open(IRPoCom::VENDOR_ID, IRPoCom::DEVICE_ID)) {
    throw IRPoComError("Device Open Error");
  }
}

void IRPoCom::SetTime(uint32_t epochTime)
{
  SendToDevice(IRPO_USB_DATA_REQ_SET_TIME, &epochTime, sizeof(uint32_t));
}

uint32_t IRPoCom::GetTime()
{
  uint32_t epochTime;
  RecvFromDevice(IRPO_USB_DATA_REQ_GET_TIME, &epochTime, sizeof(uint32_t));
  return epochTime;
}


bool IRPoCom::IsStudyDataAvailable(void)
{
  IRPO_USB_STATUS status;
  GetStatus(status);
  if (status.studyDataAvaliable) {
    return true;
  }  else {
    return false;
  }
}

void IRPoCom::GetCurrentPos(int *angle, int *tilt)
{
  IRPO_USB_STATUS status;
  GetStatus(status);
  *angle = status.angle;
  *tilt = status.tilt;
}

uint8_t IRPoCom::GetRomUpdateStatus(void)
{
  IRPO_USB_STATUS status;
  GetStatus(status);
  return status.romUpdateStatus;
}

bool IRPoCom::GetStudyData(IR_DATA *studyData)
{
  if (! IsStudyDataAvailable()) {
    return false;
  }
  RecvFromDevice(IRPO_USB_DATA_REQ_GET_STUDY_DATA, studyData, sizeof(IR_DATA));
  return true;
}

void IRPoCom::WriteIRData(const IR_DATA *irData)
{
  SendToDevice(IRPO_USB_DATA_REQ_SET_IR_DATA, irData, sizeof(IR_DATA));
}

void IRPoCom::ClearStudyData(void)
{
  uint8_t dummy = 0;
  SendToDevice(IRPO_USB_DATA_REQ_CLEAR_STUDY_DATA, &dummy, 1);
}

void IRPoCom::GetStatus(IRPO_USB_STATUS &status)
{
  IRPO_USB_DATA_PACKET packet;
  
  GetFeatureWrapper(&packet);
  if (packet.header.reqId != 0xFF) {
    throw IRPoComError("GetStatus : reqId error");
  }
  if (packet.header.seqId != 0xFF) {
    throw IRPoComError("GetStatus : seqId error");
  }

  status = *(const IRPO_USB_STATUS*)packet.data;
}

void IRPoCom::SendToDevice(int reqId, const void * data, int len)
{
    
  if (!mHid.IsOpen()) {
    throw std::logic_error("Hid is not opened");
  }
  
  int seqId = 0;
  int remainSize = len;
  const uint8_t *src = (const uint8_t*)data;
  while (remainSize != 0) {

    IRPO_USB_DATA_PACKET packet;
    packet.header.reqId = reqId;
    packet.header.seqId = seqId++;

    int payloadSize = min(remainSize, IRPO_USB_DATA_MAX_DATA_SIZE);
    memcpy(packet.data, src, payloadSize);
    remainSize -= payloadSize;
    src += payloadSize;

    SetFeatureWrapper(&packet);
  }
}

void IRPoCom::RecvFromDevice(int reqId, void * data, int len)
{

  if (!mHid.IsOpen()) {
    throw std::logic_error("Hid is not opened");
  }
  
  int seqId = 0;
  IRPO_USB_DATA_PACKET packet;
  packet.header.reqId = reqId;
  packet.header.seqId = seqId++;
  
  SetFeatureWrapper(&packet);
  
  int remainSize = len;
  uint8_t *dst = (uint8_t*)data;
  while (remainSize != 0) {
    
    GetFeatureWrapper(&packet);
    
    if (packet.header.reqId != reqId) {
      throw IRPoComError("Req Id error");
    }
    if (packet.header.seqId != seqId) {
      throw IRPoComError("Seq Id error");
    }
    seqId++;
    
    int payloadSize = min(remainSize, IRPO_USB_DATA_MAX_DATA_SIZE);
    memcpy(dst, packet.data, payloadSize);
    remainSize -= payloadSize;
    dst += payloadSize;
  }
}

void IRPoCom::SetFeatureWrapper(const void* data)
{
  uint8_t buf[64 + 1];
  buf[0] = 0;
  memcpy(buf + 1, data, 64);
  if (!mHid.SetFeature(buf, 65)) {
    throw IRPoComError("SetFeature failed");
  }
}

void IRPoCom::GetFeatureWrapper(void *data)
{
  uint8_t buf[64 + 1];
  buf[0] = 0;
  if (!mHid.GetFeature(buf, 65)) {
    throw IRPoComError("GetFeature failed");
  }
  memcpy(data, buf + 1, 64);
}

