// IRPoHost.cpp : R\[ AvP[ṼGg |Cg`܂B
//

#include "stdafx.h"
#include "IRPoCom.h"
#include <time.h>

void PrintUsage(const char* progName)
{
  printf("Usage: %s /study\n"
         "       %s /alarm HH:MM datafile.bin\n"
         "       %s /set angle tilt name datafile.bin\n",
         progName, progName, progName);
}

void AdjustDeviceTime(IRPoCom &com)
{
  _tzset();

  long timeDiffFromUTC;
  _get_timezone(&timeDiffFromUTC);

  long epochTime = (long)time(NULL);
  epochTime -= timeDiffFromUTC;
  
  com.SetTime((uint32_t)epochTime);
}

bool GetStudyData(IRPoCom &com, IR_DATA *irData)
{
  com.GetStudyData(irData);
  
  const uint8_t *p = (const uint8_t*)irData;
  uint8_t sum = std::accumulate(p, p + sizeof(IR_DATA), 0x00);
  if (sum != 0) {
    printf("New study data is broken\n");
    return false;
  }
  printf("New study data is available (Angle: %u Tilt: %u)\n", irData->angle, irData->tilt);
  return true;
}

void WriteIRDataToDevice(IRPoCom &com, const IR_DATA *irData)
{
  com.WriteIRData(irData);
  printf("IRData writing...");
  ::Sleep(3 * 1000);
  while (1) {
    
    printf(".");
    fflush(stdout);
    
    uint8_t status = com.GetRomUpdateStatus();
    switch (status) {
    case ROM_UPDATE_IDLE:
      printf(" Done\n");
      return;
    case ROM_UPDATE_INVALID_DATA:
      printf(" failed (invalid data)\n");
      return;
    case ROM_UPDATE_FAIL:
      printf(" failed (rom update)\n");
      return;
    }
    ::Sleep(500);
  }
}

void UpdateIRDataChucksum(IR_DATA *irData)
{
  irData->checksum = 0;
  const uint8_t *p = (const uint8_t*)irData;
  uint8_t sum = std::accumulate(p, p + sizeof(IR_DATA), 0x00);
  irData->checksum = (uint8_t)(256 - sum);
}

bool ReadIRDataFromFile(const char* fname, IR_DATA* irData)
{
  FILE *fp;
  if (fopen_s(&fp, fname, "rb") == 0) {
    bool readResult = false;
    if (fread(irData, sizeof(IR_DATA), 1, fp) == 1) {
      readResult = true;
    }
    fclose(fp);
    return readResult;
  } else {
    printf("cannot open %s\n", fname);
    return false;
  }
}

bool WriteIRDataToFile(const char* fname, const IR_DATA* irData)
{
  FILE *fp;
  if (fopen_s(&fp, fname, "wb") == 0) {
    fwrite(irData, sizeof(IR_DATA), 1, fp);
    fclose(fp);
    printf("The data is written to %s.\n", fname);
    return true;
  } else {
    printf("cannot open %s\n", fname);
    return false;
  }
}

bool AskYesOrNo(const char* msg, bool defaultAnswer)
{
  printf("%s [y/n]: ", msg);
  fflush(stdout);

  char c = getchar();
  getchar();
  
  bool answer = defaultAnswer;
  if ((c == 'Y') || (c == 'y')) {
    answer = true;
  } else if ((c == 'N') || (c == 'n')) {
    answer = false;
  }
  return answer;
}

void HandleStudyData(IRPoCom &com, IR_DATA *irData)
{
  char name[IR_DATA_NAME_MAX_LEN + 1];
  printf("name(max %u chars): ", IR_DATA_NAME_MAX_LEN);
  fflush(stdout);
  fgets(name, sizeof(name), stdin);
  char *n = strrchr(name, '\n');
  if (n != NULL) {
    *n = 0;
  } else {
    while(getchar() != '\n');
  }
  memcpy(irData->name, name, sizeof(name));
  irData->alarmTime = 0;
  UpdateIRDataChucksum(irData);
  
  if (AskYesOrNo("Write the data to device?", true)) {
    WriteIRDataToDevice(com, irData);
  }
  if (name[0] != 0) {
    if (AskYesOrNo("Save the data to file?", false)) {
      char fname[IR_DATA_NAME_MAX_LEN + 1 + 4];
      snprintf(fname, sizeof(fname), "%s.bin", name);
      WriteIRDataToFile(fname, irData);
    }
  }
}

void MainLoop(int argc, char* argv[])
{
  IRPoCom com;
  
  com.Open();
  AdjustDeviceTime(com);
  
  while (1) {
    int angle, tilt;

    com.GetCurrentPos(&angle, &tilt);
    printf("\rWait IR Signal... | Current Angle(0-7): %u Tilt(0-2): %u", angle, tilt);

    if (com.IsStudyDataAvailable()) {
      IR_DATA irData;
      printf("\n");
      if (GetStudyData(com, &irData)) {
        HandleStudyData(com, &irData);
      }
      com.ClearStudyData();
      printf("\n");
    }
    ::Sleep(500);
  }
}

uint32_t ConvertStringToAlarmTime(const char* strTime)
{
  uint32_t h,m;
  sscanf_s(strTime, "%d:%d", &h, &m);
  return (h << 8) | m;
}

void SetAlarm(int argc, char* argv[])
{
  if (argc < 4) {
    PrintUsage(argv[0]);
    return;
  }
  const char* strTime = argv[2];
  const char* fname = argv[3];
  
  IR_DATA irData = {0};
  if (!ReadIRDataFromFile(fname, &irData)) {
    return;
  }
  irData.alarmTime = ConvertStringToAlarmTime(strTime);

  UpdateIRDataChucksum(&irData);

  IRPoCom com;
  com.Open();
  WriteIRDataToDevice(com, &irData);
}

void SetDatafile(int argc, char* argv[])
{
  if (argc < 5) {
    PrintUsage(argv[0]);
    return;
  }
  const char* strAngle = argv[2];
  const char* strTilt = argv[3];
  const char* name = argv[4];
  const char* fname = argv[5];
  
  int angle = atoi(strAngle);
  if ((angle < 0) || (angle >= 8)) {
    printf("angle must be [0, 7]\n");
    return;
  }
  int tilt = atoi(strTilt);
  if ((tilt < 0) || (tilt >= 3)) {
    printf("tilt must be [0, 2]\n");
    return;
  }
  if (strlen(name) > IR_DATA_NAME_MAX_LEN) {
    printf("name must be within %u characters\n", IR_DATA_NAME_MAX_LEN);
    return;
  }

  
  IR_DATA irData = {0};
  if (!ReadIRDataFromFile(fname, &irData)) {
    return;
  }
  irData.angle = angle;
  irData.tilt = tilt;
  strncpy_s(irData.name, name, IR_DATA_NAME_MAX_LEN);
  irData.name[IR_DATA_NAME_MAX_LEN] = 0;
  irData.alarmTime = 0;
  
  UpdateIRDataChucksum(&irData);

  IRPoCom com;
  com.Open();
  WriteIRDataToDevice(com, &irData);
}

int main(int argc, char* argv[])
{
  const char* progName = argv[0];
  
  if (argc < 2) {
    PrintUsage(progName);
    return -1;
  }
  
  void(*func)(int,char*[]) = NULL;
  if (strcmp(argv[1], "/study") == 0) {
    func = MainLoop;
  } else if (strcmp(argv[1], "/alarm") == 0) {
    func = SetAlarm;
  } else if (strcmp(argv[1], "/set") == 0) {
    func = SetDatafile;
  } else {
    PrintUsage(progName);
    return -1;
  }
  try {
    func(argc, argv);
  }
  catch (const IRPoComError& e)
  {
    printf("Error: %s\n", e.what());
  }

  return 0;
}

