#include "AQM1248A.h"
#include "chip.h"
#include "tick_timer.h"


static void WriteCmd(uint8_t cmd)
{
  Chip_GPIO_SetPinOutLow(LPC_GPIO, 2, 2);
  Chip_SSP_SendFrame(LPC_SSP0, cmd);
  while (Chip_SSP_GetStatus(LPC_SSP0, SSP_STAT_BSY)) {
  }
}

#if !(LCD_USE_SPI_DMA)
static void WriteData(uint8_t cmd)
{
  Chip_GPIO_SetPinOutHigh(LPC_GPIO, 2, 2);
  Chip_SSP_SendFrame(LPC_SSP0, cmd);
  while (Chip_SSP_GetStatus(LPC_SSP0, SSP_STAT_BSY)) {
  }
}
#endif

static volatile bool dmaDone;
void DMA_IRQHandler(void)
{
  Chip_DMA_ClearActiveIntAChannel(LPC_DMA, DMAREQ_SSP0_TX);
  dmaDone = true;
}


#if LCD_USE_SPI_DMA
void LCDSyncMemory(const uint8_t *dispMemory)
{
  int page;
  DMA_CHDESC_T dmaDesc;
  const uint8_t *src = dispMemory;
  
  dmaDesc.xfercfg = 0;
  dmaDesc.next = 0;

  LPC_SSP0->DMACR = 0x02; // TXDMAE = 1
  for (page = 0; page < 6; page++) {
    
    WriteCmd(0xB0 + page);
    WriteCmd(0x10);
    WriteCmd(0x00);
    
    dmaDesc.source = DMA_ADDR(src + LCD_WIDTH - 1); // Last Byteを指定
    dmaDesc.dest = DMA_ADDR(&(LPC_SSP0->DR));

    dmaDone = false;
    
    Chip_GPIO_SetPinOutHigh(LPC_GPIO, 2, 2);
    Chip_DMA_SetupTranChannel(LPC_DMA, DMAREQ_SSP0_TX, &dmaDesc);
    Chip_DMA_SetupChannelTransfer(LPC_DMA, DMAREQ_SSP0_TX,
                                  (DMA_XFERCFG_CFGVALID | DMA_XFERCFG_SETINTA | DMA_XFERCFG_SWTRIG |
                                   DMA_XFERCFG_WIDTH_8 | DMA_XFERCFG_SRCINC_1 | DMA_XFERCFG_DSTINC_0 |
                                   DMA_XFERCFG_XFERCOUNT(LCD_WIDTH)));
    Chip_DMA_SetValidChannel(LPC_DMA, DMAREQ_SSP0_TX);

    while (! dmaDone) {
      __WFI();
    }

    src += LCD_WIDTH;
  }
  LPC_SSP0->DMACR = 0;
}
#else
void LCDSyncMemory(const uint8_t *dispMemory)
{
  int page;
  const uint8_t *src = dispMemory;
  
  for (page = 0; page < 6; page++) {
    
    int i;
    
    WriteCmd(0xB0 + page);
    WriteCmd(0x10);
    WriteCmd(0x00);

    for (i = 0; i < LCD_WIDTH; i++) {
      WriteData(*src++);
    }
  }
}
#endif

static void InitDevice(void)
{
  TickTimerDelayMs(50);
  WriteCmd(0xE2); // Reset
  TickTimerDelayMs(50);

  WriteCmd(0xAE); // Display Off
  WriteCmd(0xA0); // ADC = Normal
  WriteCmd(0xC8); // Common output = reverse
  WriteCmd(0xA3); // bias 1/7

  WriteCmd(0x2C); // Power Control 1
  TickTimerDelayMs(5);
  WriteCmd(0x2E); // Power Control 2
  TickTimerDelayMs(5);
  WriteCmd(0x2F); // Power Control 3
  TickTimerDelayMs(5);

  // コントラスト
  WriteCmd(0x23);
  WriteCmd(0x81);
  WriteCmd(0x1C);

  // 表示設定
  WriteCmd(0xA4);  // display all point ON/OFF = 0
  WriteCmd(0x40);  // start line = 0
  WriteCmd(0xA6);  // display normal/reverse = normal
  WriteCmd(0xAF);  // display ON
}

static void InitSPI(void)
{
  Chip_SSP_Init(LPC_SSP0);
  Chip_SSP_Set_Mode(LPC_SSP0, SSP_MODE_MASTER);
  // IDLE状態で CLK:Hi SSEL:Hi クロック立ち上げで読み込み => CPOL=1 CPHA=1
  Chip_SSP_SetFormat(LPC_SSP0, SSP_BITS_8, SSP_FRAMEFORMAT_SPI, SSP_CLOCK_CPHA1_CPOL1);
  Chip_SSP_SetBitRate(LPC_SSP0, 18 * 1000 * 1000); // 18MHz
  Chip_SSP_Enable(LPC_SSP0);
}

static void InitDMAToSPI(void)
{
  Chip_DMA_Init(LPC_DMA);
  Chip_DMA_Enable(LPC_DMA);
  Chip_DMA_SetSRAMBase(LPC_DMA, DMA_ADDR(Chip_DMA_Table));
  Chip_DMA_EnableChannel(LPC_DMA, DMAREQ_SSP0_TX);
  Chip_DMA_EnableIntChannel(LPC_DMA, DMAREQ_SSP0_TX);
  Chip_DMA_SetupChannelConfig(LPC_DMA, DMAREQ_SSP0_TX,
                              (DMA_CFG_PERIPHREQEN | DMA_CFG_TRIGBURST_SNGL | DMA_CFG_CHPRIORITY(3)));

  NVIC_EnableIRQ(DMA_IRQn);
}

void LCDInitialize(void)
{
  Chip_GPIO_Init(LPC_GPIO);
  InitSPI();
  InitDMAToSPI();
  InitDevice();
}

void LCDEnterSleepMode(void)
{
  WriteCmd(0xAC); // Static indicator off
  WriteCmd(0x00);
  WriteCmd(0xAE); // Display off
  WriteCmd(0xA5); // Display all points on
}


void LCDExitSleepMode(void)
{
  WriteCmd(0xA4); // Display all points off
  WriteCmd(0xAD); // Static indicator on
  WriteCmd(0x00);
  WriteCmd(0xAF); // Display On
}
