#include "FXOS8700CQ.h"
#include "FXOS8700CQ_regs.h"
#include <chip.h>
#include <math.h>
#include "i2c.h"
#include "tick_timer.h"

#define PI                         (float)3.14159265f

static volatile bool mTapped;

typedef struct {
  float x;
  float y;
  float z;
} VECTOR_FLOAT;

#define ZYXDR_MASK 0x08  // DR_STATUS Register and M_DR_STATUS Magnetic data-ready status register zyxdr bit mask

static uint8_t ReadReg(uint8_t reg)
{
  return I2CRegReadByte(FXOS8700CQ_ADDRESS, reg);
}

static bool ReadMultiReg(uint8_t reg, uint8_t *buf, uint8_t size)
{
  return I2CRegReadMulti(FXOS8700CQ_ADDRESS, reg, buf, size);
}

static void WriteReg(uint8_t reg, uint8_t data)
{
  I2CRegWriteByte(FXOS8700CQ_ADDRESS, reg, data);
}

static void ConvertIntVectorToFloat(const VECTOR_DATA *intVec, VECTOR_FLOAT *floatVec)
{
  floatVec->x = (float)intVec->x;
  floatVec->y = (float)intVec->y;
  floatVec->z = (float)intVec->z;
}

static bool IsConnected(void)
{
  uint8_t c = ReadReg(WHO_AM_I_REG);
  if (c == 0xC7) {
    return true;
  }
  return false;

}

static void SetStanby(void)
{
  uint8_t c = ReadReg(CTRL_REG1);
  WriteReg(CTRL_REG1, c & ~(0x01)); // Bit[0] active = 0
  TickTimerDelayMs(50);
}

static void SetActive(void)
{
  uint8_t c = ReadReg(CTRL_REG1);
  WriteReg(CTRL_REG1, c | 0x01); // Bit[0] active = 1
}

static void Init(void)
{
  SetStanby();
  
  WriteReg(XYZ_DATA_CFG_REG, 0x00);  // +/-2g range with 0.244mg/LSB

  WriteReg(CTRL_REG1, 0x00);  // HybridMode ODR 400KHz
  WriteReg(CTRL_REG2, 0x00);
  WriteReg(CTRL_REG3, 0x00);  // 割り込み関連初期状態
  WriteReg(CTRL_REG4, 0x00);  // 割り込み関連初期状態
  WriteReg(CTRL_REG5, 0x00);  // 割り込み関連初期状態
  WriteReg(M_CTRL_REG1, 0x1F | 0x80); // Hybrid Mode & Auto Calibration
  WriteReg(M_CTRL_REG2, 0x20);  // M_OUT_X_MSB register 0x33 follows the OUT_Z_LSB register 0x06 (used for burst read)
  
  
  
  SetActive();
}

bool FXOS8700CQReadAccel(VECTOR_DATA *accel)
{
  uint8_t buffer[7];
  
  if (! ReadMultiReg(STATUS_REG, buffer, 7)) {
    return false;
  }
  if (!(buffer[0] & ZYXDR_MASK)) {
    return false;
  }
  
  accel->x = (int16_t)((((int16_t)buffer[1] << 8) | buffer[2]))>> 2;
  accel->y = (int16_t)((((int16_t)buffer[3] << 8) | buffer[4]))>> 2;
  accel->z = (int16_t)((((int16_t)buffer[5] << 8) | buffer[6]))>> 2;
  
  return true;
}


bool FXOS8700CQReadMag(VECTOR_DATA *mag)
{
  uint8_t buffer[7];
  
  if (! ReadMultiReg(M_DR_STATUS_REG, buffer, 7)) {
    return false;
  }
  if (!(buffer[0] & ZYXDR_MASK)) {
    return false;
  }

  mag->x = ((int16_t)buffer[1] << 8) | buffer[2];
  mag->y = ((int16_t)buffer[3] << 8) | buffer[4];
  mag->z = ((int16_t)buffer[5] << 8) | buffer[6];

  return true;
}

bool FXOS8700CQReadAccessAndMag(VECTOR_DATA *accel, VECTOR_DATA *mag)
{
  uint8_t buffer[13];
  if (! ReadMultiReg(STATUS_REG, buffer, 13)) {
    return false;
  }
  if (!(buffer[0] & ZYXDR_MASK)) {
    return false;
  }

  accel->x = (int16_t)(((buffer[1] << 8) | buffer[2]))>> 2;
  accel->y = (int16_t)(((buffer[3] << 8) | buffer[4]))>> 2;
  accel->z = (int16_t)(((buffer[5] << 8) | buffer[6]))>> 2;
  
  mag->x = (buffer[7] << 8) | buffer[8];
  mag->y = (buffer[9] << 8) | buffer[10];
  mag->z = (buffer[11] << 8) | buffer[12];

  return true;
}

bool FXOS8700CQReadAngle(ANGLE_DATA *angle)
{
  VECTOR_DATA accelIntVec, magIntVec;
  VECTOR_FLOAT accel, mag;
  float accelNorm;
  float cosX, cosY, sinX, sinY;
  float mx, my;
  float heading;

  if (! FXOS8700CQReadAccessAndMag(&accelIntVec, &magIntVec)) {
    return false;
  }
  ConvertIntVectorToFloat(&accelIntVec, &accel);
  ConvertIntVectorToFloat(&magIntVec, &mag);
  
  accelNorm = sqrt(accel.x * accel.x + accel.y * accel.y + accel.z * accel.z);
  angle->pitch = -asin(accel.x / accelNorm);
  angle->roll = -atan2(accel.y, accel.z);

  cosY = cos(angle->pitch);
  sinY = sin(angle->pitch);
  cosX = cos(-angle->roll);
  sinX = sin(-angle->roll);
  
  mx = cosY * mag.x + sinY * sinX * mag.y + sinY * cosX * mag.z;
  my = cosX * mag.y - sinX * mag.z;
  heading = atan2(-my, -mx);
  
  while (heading < 0) {
    heading = heading + 2.0f*PI;
  }
  while (2.0f*PI <= heading) {
    heading = heading - 2.0f*PI;
  }
  angle->heading = heading;
  
  return true;
}

void GINT0_IRQHandler(void)
{
  Chip_GPIOGP_ClearIntStatus(LPC_GPIOGROUP, 0);
  mTapped = true;
}

bool FXOS8700CQEnableTapDetect(void)
{

  mTapped = false;
  
  SetStanby();

  // PULSE割り込み設定
  WriteReg(PULSE_CFG_REG, 0x1 << 4); // z-axis single tap
  WriteReg(PULSE_THSZ_REG, 0x2A);
  WriteReg(PULSE_TMLT_REG, 0x50);
  WriteReg(PULSE_LTCY_REG, 0xF0);

  // PULSE割り込みでINT2をLowにするように設定
  WriteReg(CTRL_REG3, 0x00);
  WriteReg(CTRL_REG4, 0x01 << 3);
  WriteReg(CTRL_REG5, 0x00);

  // CPU側 PIO1_13 Low でGroup Int 0割り込みが発生するように設定
  Chip_Clock_EnablePeriphClock(SYSCTL_CLOCK_GROUP0INT);
  Chip_GPIOGP_SelectEdgeMode(LPC_GPIOGROUP, 0);
  Chip_GPIOGP_SelectLowLevel(LPC_GPIOGROUP, 0, 1, 1 << 13);
  Chip_GPIOGP_EnableGroupPins(LPC_GPIOGROUP, 0, 1, 1 << 13);
  NVIC_EnableIRQ(GINT0_IRQn);

  SetActive();

  return true;
}

bool FXOS8700CQIsTapDetected(void)
{
  bool ret = mTapped;
  mTapped = false;
  return ret;
}

bool FXOS8700CQSetupSleepWakeup(void)
{
  SetStanby();

  // PULSE割り込みはもはや使わないので割り込みを無視するようにする
  NVIC_DisableIRQ(GINT0_IRQn);
  Chip_GPIOGP_DisableGroupPins(LPC_GPIOGROUP, 0, 1, 1 << 13);
  
  // TODO 本来はSleepモードを活用すべきだが、面倒なので単純にMag vectorでの割り込みを利用する
  WriteReg(A_VECM_THS_MSB_REG, 0x80 | 0x07); // Clear debounce counter when event falls below threshold; bit 4:0 MSB THS
  WriteReg(A_VECM_THS_LSB_REG, 0xFF);        // Specify threshold at 0.3 g = 1228 * 0.244 g at FSR of +/-2 g
  WriteReg(A_VECM_CNT_REG, 0x01);
  WriteReg(A_VECM_CFG_REG, 1 << 3);
  
  // Mag VectorでINT0(INT0はOpenDrain)をLowにするように設定
  WriteReg(CTRL_REG3, 0x01 << 0);
  WriteReg(CTRL_REG4, 0x01 << 1);
  WriteReg(CTRL_REG5, 0x01 << 1);
  
  SetActive();

  return true;
}

bool InitializeFXOS8700CQ(void)
{
  InitializeI2C();

  if (! IsConnected()) {
    return false;
  }

  Init();
  
  return true;
}


