#include "IRLed.h"
#include <chip.h>
#include <string.h>
#include "tick_timer.h"
#include "IRData.h"
#include "IRPoDisplay.h"

extern int mCurrentAngle;
extern int mCurrentTilt;

static const int8_t CAPTURE_CHANNEL = 0;
static const uint32_t CAPTURE_TIMER_FREQ  = 100 * 1000; // 100KHz (10us)

#define IR_FREQ 38000
#define STUDY_TURN_OFF_COUNT   (500 * 1000 / 10)    // 500msec分のカウント

typedef struct {
  uint32_t sampleCount;
  uint32_t sumValue;
  uint32_t rangeLow;
  uint32_t rangeHi;
} TIME_GROUP;

typedef struct {
  uint32_t sampleCount;
  TIME_GROUP timeGroup[IR_DATA_TIME_GROUP_MAX_COUNT];
  uint8_t samples[IR_DATA_TIMES_MAX_COUNT];
} IR_STUDY_DATA;

typedef enum {
  IR_STUDY_IDLE,
  IR_STUDY_RUN,
  IR_STUDY_DONE,
  IR_STUDY_WAIT_CLEAR,
  IR_STUDY_ERR,
} IR_STUDY_STATUS;

static IR_STUDY_DATA mStudyData;
static IR_DATA mIRData;
static volatile IR_STUDY_STATUS mStudyStatus;
static volatile uint32_t mPrevCounter;

static bool RecordNewSample(IR_STUDY_DATA *data, uint32_t sample)
{
  int groupIndex;
  TIME_GROUP *group = data->timeGroup;
  
  if (data->sampleCount >= IR_DATA_TIMES_MAX_COUNT) {
    return false;
  }
  
  for (groupIndex = 0; groupIndex < IR_DATA_TIME_GROUP_MAX_COUNT; groupIndex++) {
    if (group->sampleCount == 0) {
      uint32_t margin = (sample / 10) + 1;
      group->rangeLow = sample - margin;
      group->rangeHi = sample + margin;
      break;
    }
    if ((group->rangeLow <= sample) && (sample < group->rangeHi)) {
      break;
    }
    group++;
  }
  if (groupIndex == IR_DATA_TIME_GROUP_MAX_COUNT) {
    return false; // オーバーフロー
  }
  group->sampleCount++;
  group->sumValue += sample;
  
  data->samples[data->sampleCount] = groupIndex;
  data->sampleCount++;

  return true;
}

void TIMER32_0_IRQHandler(void)
{
  if (Chip_TIMER_CapturePending(LPC_TIMER32_0, CAPTURE_CHANNEL)) {
    
    uint32_t counter;
    
    Chip_TIMER_ClearCapture(LPC_TIMER32_0, CAPTURE_CHANNEL);
    counter = Chip_TIMER_ReadCapture(LPC_TIMER32_0, CAPTURE_CHANNEL);
    
    switch (mStudyStatus) {
    case IR_STUDY_IDLE:
      mStudyStatus = IR_STUDY_RUN;
      break;
    case IR_STUDY_RUN:
      {
        uint32_t diff = counter - mPrevCounter;
        if (diff > STUDY_TURN_OFF_COUNT) {
          mStudyStatus = IR_STUDY_DONE;
        } else if (diff != 0) {
          if (! RecordNewSample(&mStudyData, diff)) {
            mStudyStatus = IR_STUDY_ERR;
          }
        }
      }
      break;
    default:
      // DoNothing
      break;
    }
    mPrevCounter = counter;
  }
}

static void ConvertStudyDataToIRData(const IR_STUDY_DATA* studyData, int angle, int tilt, IR_DATA *runData)
{
  int groupIndex;
  const TIME_GROUP *group = studyData->timeGroup;

  memset(runData, 0, sizeof(IR_DATA));
  runData->angle = angle;
  runData->tilt = tilt;
  
  for (groupIndex = 0; groupIndex < IR_DATA_TIME_GROUP_MAX_COUNT; groupIndex++) {
    uint32_t len;
    if (group->sampleCount == 0) {
      break;
    }
    len = (group->sumValue * 100 / group->sampleCount) / 263;
    if (len == 0) {
      len = 1;
    }
    runData->timeGroup[groupIndex] = len;
    group++;
  }
  runData->timesCount = studyData->sampleCount;
  memcpy(runData->times, &studyData->samples, studyData->sampleCount);
  
  runData->checksum = (uint8_t)(256 - IRDataCalcChecksum(runData));
}

void IRLedStudyProcess(void)
{
  
  NVIC_DisableIRQ(TIMER_32_0_IRQn);
  if (mStudyStatus == IR_STUDY_RUN) {
    uint32_t diff = Chip_TIMER_ReadCount(LPC_TIMER32_0) - mPrevCounter;
    if (diff > STUDY_TURN_OFF_COUNT) {
      mStudyStatus = IR_STUDY_DONE;
    }
  }
  NVIC_EnableIRQ(TIMER_32_0_IRQn);
  
  switch (mStudyStatus) {
  case IR_STUDY_DONE:
    IRPoDisplayUpdate(mCurrentAngle, mCurrentTilt);
    ConvertStudyDataToIRData(&mStudyData, mCurrentAngle, mCurrentTilt, &mIRData);
    mStudyStatus = IR_STUDY_WAIT_CLEAR;
    break;
  case IR_STUDY_ERR:
    IRLedClearStudyData();
    break;
  default:
    // Do Nothing
    break;
  }
}

const IR_DATA* IRLedGetStudyData(void)
{
  if (mStudyStatus == IR_STUDY_WAIT_CLEAR) {
    return &mIRData;
  } else {
    return NULL;
  }
}

void IRLedClearStudyData(void)
{
  NVIC_DisableIRQ(TIMER_32_0_IRQn);
  
  memset(&mStudyData, 0, sizeof(mStudyData));
  mStudyStatus = IR_STUDY_IDLE;
  mPrevCounter = Chip_TIMER_ReadCount(LPC_TIMER32_0);
  
  NVIC_EnableIRQ(TIMER_32_0_IRQn);
}

void IRLedStudyInitialize(void)
{
  uint32_t systemFreq;

  // 赤外線受信モジュール ON
  Chip_GPIO_SetPinOutHigh(LPC_GPIO, 0, 7);
  TickTimerDelayMs(500);

  // 受信モジュールの信号変化タイミング測定のためのタイマ設定&起動
  //  - タイマの周波数は 100KHz (10us) とする
  //  - IR_INの両エッジでcounterをcaptureす
  Chip_TIMER_Init(LPC_TIMER32_0);
  
  systemFreq = Chip_Clock_GetSystemClockRate();
  
  Chip_TIMER_Reset(LPC_TIMER32_0);
  Chip_TIMER_PrescaleSet(LPC_TIMER32_0, (systemFreq / CAPTURE_TIMER_FREQ) - 1);
  
  Chip_TIMER_CaptureEnableInt(LPC_TIMER32_0, CAPTURE_CHANNEL);
  Chip_TIMER_CaptureFallingEdgeEnable(LPC_TIMER32_0, CAPTURE_CHANNEL);
  Chip_TIMER_CaptureRisingEdgeEnable(LPC_TIMER32_0, CAPTURE_CHANNEL);
  
  NVIC_ClearPendingIRQ(TIMER_32_0_IRQn);
  NVIC_EnableIRQ(TIMER_32_0_IRQn);
  
  Chip_TIMER_Enable(LPC_TIMER32_0);
  
  IRLedClearStudyData();
}

void IRLedStudyUninitalize(void)
{
  // タイマー停止
  NVIC_DisableIRQ(TIMER_32_0_IRQn);
  Chip_TIMER_Disable(LPC_TIMER32_0);
  
  // 赤外線受信モジュールOff
  Chip_GPIO_SetPinOutLow(LPC_GPIO, 0, 7);
}

static volatile bool mIROutput;
static volatile uint32_t mCurrentIndex;
static volatile uint32_t mCurrentTick;
static volatile uint32_t mNextTick;
static const IR_DATA *mCurrentData;

void TIMER32_1_IRQHandler(void)
{
  if (Chip_TIMER_MatchPending(LPC_TIMER32_1, 1)) {
    Chip_TIMER_ClearMatch(LPC_TIMER32_1, 1);
    
    mCurrentTick++;
    if (mCurrentTick == mNextTick) {
      
      mCurrentTick = 0;
      
      if (mIROutput) {
        LPC_TIMER32_1->PWMC = 0;
        mIROutput = false;
      } else {
        LPC_TIMER32_1->PWMC = 1; // PWMEND0 (Bit[0]) = 1
        mIROutput = true;
      }
      
      mCurrentIndex++;
      if (mCurrentIndex < mCurrentData->timesCount) {
        uint8_t data = mCurrentData->times[mCurrentIndex];
        mNextTick = mCurrentData->timeGroup[data];
      } else {
        Chip_TIMER_Disable(LPC_TIMER32_1);
      }
    }
  }
}

void IRLedSend(const IR_DATA *data)
{
  uint32_t systemFreq;

  systemFreq = Chip_Clock_GetSystemClockRate();
  
  Chip_TIMER_Init(LPC_TIMER32_1);
  
  Chip_TIMER_Reset(LPC_TIMER32_1);
  Chip_TIMER_PrescaleSet(LPC_TIMER32_1, (systemFreq / (IR_FREQ * 10)) - 1);
  
  Chip_TIMER_SetMatch(LPC_TIMER32_1, 0, 7);
  
  Chip_TIMER_SetMatch(LPC_TIMER32_1, 1, 9);
  Chip_TIMER_ResetOnMatchEnable(LPC_TIMER32_1, 1);
  Chip_TIMER_MatchEnableInt(LPC_TIMER32_1, 1);
  
  NVIC_ClearPendingIRQ(TIMER_32_1_IRQn);
  NVIC_EnableIRQ(TIMER_32_1_IRQn);
  
  mCurrentData = data;
  mIROutput = true;
  mCurrentIndex = 0;
  mCurrentTick = 0;
  mNextTick = data->timeGroup[data->times[0]];
  
  LPC_TIMER32_1->PWMC = 1; // PWMEND0 (Bit[0]) = 1
  Chip_TIMER_Enable(LPC_TIMER32_1);
  
  while (mCurrentIndex < mCurrentData->timesCount) {
    __WFI();
  }
}
