#include "chip.h"
#include <cr_section_macros.h>
#include <stdio.h>
#include <time.h>
#include "tick_timer.h"
#include "rtc.h"
#include "IRPoDisplay.h"
#include "IRPoUsbComm.h"
#include "IRLed.h"
#include "IRData.h"
#include "FXOS8700CQ.h"

#define PI                         (float)3.14159265f
int mCurrentAngle = 0;
int mCurrentTilt = 0;

#define SLEEP_INTERVAL_MS   10000

void UpdateAngleTilt(void)
{
  ANGLE_DATA sensorValue;
  float heading, pitch;

  while (! FXOS8700CQReadAngle(&sensorValue));
  
  heading = sensorValue.heading * 180.0f / PI;
  if (heading >= 360.0f) {
    heading -= 360.0f;
  }
  mCurrentAngle = (int)(heading / 45);

  pitch = sensorValue.pitch * 180.0f/ PI;
  if (pitch < -20) {
    mCurrentTilt = 2;
  } else if (pitch < 20) {
    mCurrentTilt = 1;
  } else {
    mCurrentTilt = 0;
  }
}

static void EnterDeepPowerDown(void)
{
  IRPoDisplayOff();
  FXOS8700CQSetupSleepWakeup();
  Chip_PMU_DeepPowerDownState(LPC_PMU);
}

static void StudyMain(void)
{
  uint16_t prevTick;

  IRLedStudyInitialize();

  UpdateAngleTilt();
  
  IRPoDisplaySetStatusString("Study Mode");
  IRPoDisplayUpdate(mCurrentAngle, mCurrentTilt);
  
  prevTick = Chip_TIMER_ReadCount(LPC_TIMER16_0);
  
  while (IRPoIsUsbConnected()) {
    {
      uint16_t curTick, diffTick;
      curTick = Chip_TIMER_ReadCount(LPC_TIMER16_0);
      diffTick = curTick - prevTick;
      if (diffTick > 250) {
        UpdateAngleTilt();
        IRPoDisplayUpdate(mCurrentAngle, mCurrentTilt);
        prevTick = curTick;
      }
    }
    IRLedStudyProcess();
    IRPoUsbCommProcess();
  }
}

static void ControllerMain(void)
{
  uint16_t prevTick;
  uint16_t prevTickForSleep;
  
  UpdateAngleTilt();
  
  IRPoDisplaySetStatusString("Remote Mode");
  IRPoDisplayUpdate(mCurrentAngle, mCurrentTilt);

  FXOS8700CQEnableTapDetect();
  
  prevTick = Chip_TIMER_ReadCount(LPC_TIMER16_0);
  prevTickForSleep = Chip_TIMER_ReadCount(LPC_TIMER16_0);
  
  while (!IRPoIsUsbConnected()) {

    {
      uint16_t curTick, diffTick;
      curTick = Chip_TIMER_ReadCount(LPC_TIMER16_0);
      diffTick = curTick - prevTick;
      if (diffTick > 250) {
        UpdateAngleTilt();
        IRPoDisplayUpdate(mCurrentAngle, mCurrentTilt);
        prevTick = curTick;
      }
    }
    
    if (FXOS8700CQIsTapDetected()) {
      
      const IR_DATA *data = IRDataGetPointer(AngleTiltToIndex(mCurrentAngle, mCurrentTilt));
      if ((data->timesCount > 0) && (data->timesCount <= IR_DATA_TIMES_MAX_COUNT)) {
        
        IRPoDisplaySetStatusString("Send Cmd");
        IRLedSend(data);
        TickTimerDelayMs(300);
        IRPoDisplaySetStatusString("Remote Mode");
      }

      prevTickForSleep = Chip_TIMER_ReadCount(LPC_TIMER16_0);
    }
    
    {
      uint16_t curTick, diffTick;
      curTick = Chip_TIMER_ReadCount(LPC_TIMER16_0);
      diffTick = curTick - prevTickForSleep;
      if (diffTick > SLEEP_INTERVAL_MS) {
        EnterDeepPowerDown();
      }
    }
  }
}

static void SendAlarmedSignal(void)
{
  const IR_DATA *data = IRDataGetPointer(IR_DATA_ALARM_INDEX);
  if ((data->timesCount > 0) && (data->timesCount <= IR_DATA_TIMES_MAX_COUNT)) {
    
    UpdateAngleTilt();
    
    IRPoDisplaySetStatusString("- ALARMED - ");
    IRPoDisplayUpdate(mCurrentAngle, mCurrentTilt);
    
    IRLedSend(data);
    
    TickTimerDelayMs(500);
    TickTimerDelayMs(500);
    TickTimerDelayMs(500);
  }
}

static void SetNextAlarmTimeIfNeeded(void)
{
  const IR_DATA *data = IRDataGetPointer(IR_DATA_ALARM_INDEX);
  if ((data->timesCount > 0) && (data->timesCount <= IR_DATA_TIMES_MAX_COUNT)) {

    time_t currentTime, dayTime, offsetTime;
    struct tm *tmStruct;
    int alarmH, alarmM;
    int alarmOffset;
    
    currentTime = time(NULL);
    tmStruct = localtime(&currentTime);
    offsetTime = (((tmStruct->tm_hour * 60) + tmStruct->tm_min) * 60) + tmStruct->tm_sec;
    dayTime = currentTime - offsetTime;
    
    alarmH = (uint8_t)(data->alarmTime >> 8);
    alarmM = (uint8_t)(data->alarmTime);
    alarmOffset = (alarmH * 60 + alarmM) * 60;
    
    if (offsetTime > alarmOffset) {
      dayTime += 24 * 60 * 60;
    }
    
    RTCSetAlarmByEpoch(dayTime + alarmOffset);
  }
}

int main(void) {

  while (1) {
    bool alarmed = false;
    
    SystemCoreClockUpdate();
    InitTickTimer();
    RTCInitialize(&alarmed);
    SetNextAlarmTimeIfNeeded();
    
    InitializeFXOS8700CQ();
    
    IRPoUsbCommInitialize();
    IRPoDisplayInitialize();
    
    while (1) {
      if (IRPoIsUsbConnected()) {
        StudyMain();
      } else {
        if (alarmed) {
          SendAlarmedSignal();
        }
        ControllerMain();
      }
    }
  }
  return 0 ;
}

