#include "usb_hid_comm.h"
#include <usbd/usbd_hid.h>
#include <chip.h>
#include <string.h>
#include <time.h>
#include "IRPoUsbData.h"
#include "rtc.h"
#include "IRLed.h"

extern int mCurrentAngle;
extern int mCurrentTilt;

#define IRPO_VENDOR_ID  0x6666   // Prototype
#define IRPO_PRODUCT_ID 0x2844

#define HID_INOUT_BYTES  64

#define USB_SRAM_BASE 0x20004000
#define USB_SRAM_SIZE 0x800 // 2KB


static const uint8_t mReportDesc[] = {
  HID_UsagePageVendor(0x00),
  HID_Usage(0x01),
  HID_Collection(HID_Application),
  HID_LogicalMin(0),	/* value range: 0 - 0xFF */
  HID_LogicalMaxS(0xFF),
  HID_ReportSize(8),	/* 8 bits */
  HID_ReportCount(HID_INOUT_BYTES),
  HID_Usage(0x01),
  HID_Feature(HID_Data | HID_Variable | HID_Absolute),
  HID_EndCollection,
};

static const uint8_t mDeviceDesc[] = {
  USB_DEVICE_DESC_SIZE,         /* bLength */
  USB_DEVICE_DESCRIPTOR_TYPE,   /* bDescriptorType */
  WBVAL(0x0200),                /* bcdUSB 2.0 */
  0x00,                         /* bDeviceClass */
  0x00,                         /* bDeviceSubClass */
  0x00,                         /* bDeviceProtocol */
  USB_MAX_PACKET0,              /* bMaxPacketSize0 */
  WBVAL(IRPO_VENDOR_ID),        /* idVendor */
  WBVAL(IRPO_PRODUCT_ID),       /* idProduct */
  WBVAL(0x0100),                /* bcdDevice */
  0x01,                         /* iManufacturer */
  0x02,                         /* iProduct */
  0x00,                         /* iSerialNumber */
  0x01                          /* bNumConfigurations */
  };


/**
 * USB FSConfiguration Descriptor
 * All Descriptors (Configuration, Interface, Endpoint, Class, Vendor)
 */
static const uint8_t mConfigDesc[] = {
  
  /* Configuration 1 */
  USB_CONFIGURATION_DESC_SIZE,            /* bLength */
  USB_CONFIGURATION_DESCRIPTOR_TYPE,      /* bDescriptorType */
  WBVAL(                                  /* wTotalLength */
    USB_CONFIGURATION_DESC_SIZE   +
    USB_INTERFACE_DESC_SIZE       +
    HID_DESC_SIZE),
  0x01,                           /* bNumInterfaces */
  0x01,                           /* bConfigurationValue */
  0x00,                           /* iConfiguration */
  USB_CONFIG_SELF_POWERED,        /* bmAttributes */
  USB_CONFIG_POWER_MA(250),       /* bMaxPower */
  
  /* Interface 0, Alternate Setting 0, HID Class */
  USB_INTERFACE_DESC_SIZE,        /* bLength */
  USB_INTERFACE_DESCRIPTOR_TYPE,  /* bDescriptorType */
  0x00,                           /* bInterfaceNumber */
  0x00,                           /* bAlternateSetting */
  0x00,                           /* bNumEndpoints */
  USB_DEVICE_CLASS_HUMAN_INTERFACE,   /* bInterfaceClass */
  HID_SUBCLASS_NONE,              /* bInterfaceSubClass */
  HID_PROTOCOL_NONE,              /* bInterfaceProtocol */
  0x00,                           /* iInterface */
  
  /* HID Class Descriptor */
  HID_DESC_SIZE,                  /* bLength */
  HID_HID_DESCRIPTOR_TYPE,        /* bDescriptorType */
  WBVAL(0x0111),                  /* bcdHID : 1.11*/
  0x00,                           /* bCountryCode */
  0x01,                           /* bNumDescriptors */
  HID_REPORT_DESCRIPTOR_TYPE,     /* bDescriptorType */
  WBVAL(sizeof(mReportDesc)),     /* wDescriptorLength */
  /* Terminator */
  0                               /* bLength */
  };

/**
 * USB String Descriptor (optional)
 */
static const uint8_t mStringDesc[] = {
  /* Index 0x00: LANGID Codes */
  0x04,                           /* bLength */
  USB_STRING_DESCRIPTOR_TYPE,     /* bDescriptorType */
  WBVAL(0x0409),                  /* wLANGID : US English*/
  /* Index 0x01: Manufacturer */
  (9 * 2 + 2),                   /* bLength (8 Char + Type + length) */
  USB_STRING_DESCRIPTOR_TYPE,     /* bDescriptorType */
  'd', 0,
  '-', 0,
  'r', 0,
  'i', 0,
  's', 0,
  's', 0,
  'o', 0,
  'k', 0,
  'u', 0,
  /* Index 0x02: Product */
  (4 * 2 + 2),                   /* bLength (4 Char + Type + length) */
  USB_STRING_DESCRIPTOR_TYPE,     /* bDescriptorType */
  'I', 0,
  'R', 0,
  'P', 0,
  'o', 0,
};

static USB_HID_REPORT_T mReportData = {.len = sizeof(mReportDesc), .idle_time = 0, .desc = (uint8_t*)mReportDesc};

static IR_DATA mWriteData;
static uint8_t mRomUpdateStatus;

static uint8_t mToHostReqId = 0xFF;
static uint8_t mToHostSeqId = 0;
static const uint8_t *mToHostData = NULL;
static uint16_t mToHostDataRemainSize = 0;

static uint8_t mFromHostReqId = 0xFF;
static uint8_t mFromHostSeqId = 0xFF;

static void SendToHostHandler(uint8_t *data, uint16_t *length)
{
  IRPO_USB_DATA_PACKET *packet = (IRPO_USB_DATA_PACKET*)data;
  
  packet->header.reqId = mToHostReqId;
  if (mToHostDataRemainSize == 0) {
    packet->header.seqId = 0xFF;
    if (mToHostReqId == 0xFF) {
      
      IRPO_USB_STATUS *status = (IRPO_USB_STATUS*)packet->data;
      if (IRLedGetStudyData() != NULL) {
        status->studyDataAvaliable = true;
      } else {
        status->studyDataAvaliable = false;
      }
      status->angle = mCurrentAngle;
      status->tilt = mCurrentTilt;
      status->romUpdateStatus = mRomUpdateStatus;
    }
  } else {
    uint8_t payloadSize;
    
    packet->header.seqId = mToHostSeqId;
    mToHostSeqId++;

    if (mToHostDataRemainSize > IRPO_USB_DATA_MAX_DATA_SIZE) {
      payloadSize = IRPO_USB_DATA_MAX_DATA_SIZE;
    } else {
      payloadSize = mToHostDataRemainSize;
    }
    memcpy(packet->data, mToHostData, payloadSize);
    mToHostDataRemainSize -= payloadSize;
    mToHostData += payloadSize;

    if (mToHostDataRemainSize == 0) {
      mToHostReqId = 0xFF;
    }
  }
  *length = sizeof(IRPO_USB_DATA_PACKET);
}

static void HandleGetTime(const IRPO_USB_DATA_PACKET *packet)
{
  static uint32_t epochTime = 0;
  epochTime = time(NULL);
  mToHostReqId = packet->header.reqId;
  mToHostSeqId = packet->header.seqId + 1;
  mToHostData = (const uint8_t*)&epochTime;
  mToHostDataRemainSize = sizeof(uint32_t);
}

static void HandleSetTime(const IRPO_USB_DATA_PACKET *packet)
{
  uint32_t epochTime;
  memcpy(&epochTime, packet->data, sizeof(uint32_t));
  RTCSetTimeByEpoch(epochTime);
}

static void HandleGetStudyData(const IRPO_USB_DATA_PACKET *packet)
{
  const IR_DATA *studyData;

  studyData = IRLedGetStudyData();
  if (studyData == NULL) {
    return;
  }
  
  mToHostReqId = packet->header.reqId;
  mToHostSeqId = packet->header.seqId + 1;
  mToHostData = (const uint8_t*)studyData;
  mToHostDataRemainSize = sizeof(IR_DATA);
}

static void HandleSetIrData(const IRPO_USB_DATA_PACKET *packet)
{
  static uint8_t *dst;
  static int remainSize = 0;
  
  if (packet->header.seqId == 0) {
    dst = (uint8_t*)&mWriteData;
    remainSize = sizeof(IR_DATA);
    mRomUpdateStatus = ROM_UPDATE_RECVING;
  }
  if (remainSize != 0) {
    int copySize;
    if (remainSize > IRPO_USB_DATA_MAX_DATA_SIZE) {
      copySize = IRPO_USB_DATA_MAX_DATA_SIZE;
    } else {
      copySize = remainSize;
    }
    memcpy(dst, packet->data, copySize);
    remainSize -= copySize;
    dst += copySize;
  }
  if (remainSize != 0) {
    mFromHostReqId = packet->header.reqId;
    mFromHostSeqId = packet->header.seqId + 1;
  } else {
    if (IRDataCalcChecksum(&mWriteData) == 0) {
      mRomUpdateStatus = ROM_UPDATE_REQUESTING;
    } else {
      mRomUpdateStatus = ROM_UPDATE_INVALID_DATA;
    }
    mFromHostReqId = 0xFF;
    mFromHostSeqId = 0xFF;
  }
}
                            
static void HandleClearStudyData(const IRPO_USB_DATA_PACKET *packet)
{
  IRLedClearStudyData();
}

static void RecvFromHostHandler(const uint8_t *data, uint16_t length)
{
  const IRPO_USB_DATA_PACKET *packet = (const IRPO_USB_DATA_PACKET*)data;

  if (length == 0) {
    return;
  }
  
  if (packet->header.seqId == 0) {
    switch (packet->header.reqId) {
    case IRPO_USB_DATA_REQ_GET_TIME:
      HandleGetTime(packet);
      break;
    case IRPO_USB_DATA_REQ_SET_TIME:
      HandleSetTime(packet);
      break;
    case IRPO_USB_DATA_REQ_GET_STUDY_DATA:
      HandleGetStudyData(packet);
      break;
    case IRPO_USB_DATA_REQ_SET_IR_DATA:
      HandleSetIrData(packet);
      break;
    case IRPO_USB_DATA_REQ_CLEAR_STUDY_DATA:
      HandleClearStudyData(packet);
      break;
    default:
      // DoNothing
      break;
    }
  } else {
    if ((packet->header.reqId != mFromHostReqId) || (packet->header.seqId != mFromHostSeqId)) {
      mFromHostReqId = 0xFF;
      mFromHostSeqId = 0xFF;
      return;
    }
    switch (packet->header.reqId) {
    case IRPO_USB_DATA_REQ_SET_IR_DATA:
      HandleSetIrData(packet);
    default:
      // Do Nothing
      break;
    }
  }
}

bool IRPoIsUsbConnected(void)
{
  return (LPC_USB->DEVCMDSTAT & (1UL << 28));  // Bit28: VBUSDEBOUNCED
}

void IRPoUsbCommProcess(void)
{
  if (mRomUpdateStatus == ROM_UPDATE_REQUESTING) {
    int index;
    
    mRomUpdateStatus = ROM_UPDATE_RUNNING;

    if (mWriteData.alarmTime == 0) {
      index = AngleTiltToIndex(mWriteData.angle, mWriteData.tilt);
    } else {
      index = IR_DATA_ALARM_INDEX;
    }
    if (IRDataWriteToFlash(index, &mWriteData)) {
      mRomUpdateStatus = ROM_UPDATE_IDLE;
    } else {
      mRomUpdateStatus = ROM_UPDATE_FAIL;
    }
  }
}

void IRPoUsbCommInitialize(void)
{
  USB_HID_COMM_PARAM commParam = {0};

  commParam.memBase = USB_SRAM_BASE;
  commParam.memSize = USB_SRAM_SIZE;
  commParam.deviceDesc = mDeviceDesc;
  commParam.stringDesc = mStringDesc;
  commParam.configDesc = mConfigDesc;
  commParam.reportData = &mReportData;
  commParam.sendToHostHandler = SendToHostHandler;
  commParam.recvFromHostHandler = RecvFromHostHandler;
  
  USBHidCommInitialize(&commParam);
  
}
