#include "usb_hid_comm.h"
#include <chip.h>
#include <usbd/usbd.h>
#include <usbd/usbd_core.h>
#include <usbd/usbd_rom_api.h>

const  USBD_API_T *g_pUsbApi;
static USBD_HANDLE_T mUsbHandle;
static USB_HID_COMM_SEND_TO_HOST_HANDLER mSendToHostHandler;
static USB_HID_COMM_RECV_FROM_HOST_HANDLER mRecvFromHostHandler;

typedef volatile struct _EP_LIST {
  uint32_t buf_ptr;
  uint32_t buf_length;
} EP_LIST;

static ErrorCode_t StallWorkAround(USBD_HANDLE_T hUsb)
{
  ErrorCode_t ret = LPC_OK;
  
  USB_CORE_CTRL_T *pCtrl = (USB_CORE_CTRL_T *) hUsb;
  EP_LIST *epQueue;
  int32_t i;
  
  /* WORKAROUND for Case 2:
  Code clearing STALL bits in endpoint reset routine corrupts memory area
  next to the endpoint control data.
  */
  if (pCtrl->ep_halt != 0) { /* check if STALL is set for any endpoint */
    /* get pointer to HW EP queue */
    epQueue = (EP_LIST *) LPC_USB->EPLISTSTART;
    /* check if the HW STALL bit for the endpoint is cleared due to bug. */
    for (i = 1; i < pCtrl->max_num_ep; i++) {
      /* check OUT EPs */
      if ( pCtrl->ep_halt & (1 << i)) {
        /* Check if HW EP queue also has STALL bit = _BIT(29) is set */
        if (( epQueue[i << 1].buf_ptr & _BIT(29)) == 0) {
          /* bit not set, cleared by BUG. So set it back. */
          epQueue[i << 1].buf_ptr |= _BIT(29);
        }
      }
      /* Check IN EPs */
      if ( pCtrl->ep_halt & (1 << (i + 16))) {
        /* Check if HW EP queue also has STALL bit = _BIT(29) is set */
        if (( epQueue[(i << 1) + 1].buf_ptr & _BIT(29)) == 0) {
          /* bit not set, cleared by BUG. So set it back. */
          epQueue[(i << 1) + 1].buf_ptr |= _BIT(29);
        }
      }
    }
  }
  return ret;
}

static ErrorCode_t GetReport(USBD_HANDLE_T hHid, USB_SETUP_PACKET *pSetup, uint8_t * *pBuffer, uint16_t *plength)
{
  uint8_t reportType = pSetup->wValue.WB.H;
  switch(reportType) {
  case HID_REPORT_FEATURE:
    mSendToHostHandler(*pBuffer, plength);
    break;
  default:
    return ERR_USBD_STALL; // 未サポート
  }
  return LPC_OK;
}

static ErrorCode_t SetReport(USBD_HANDLE_T hHid, USB_SETUP_PACKET *pSetup, uint8_t * *pBuffer, uint16_t length)
{
  uint8_t reportType = pSetup->wValue.WB.H;
  switch(reportType) {
  case HID_REPORT_FEATURE:
    mRecvFromHostHandler(*pBuffer, length);
    break;
  default:
    return ERR_USBD_STALL;
  }
  return LPC_OK;
}

static const USB_INTERFACE_DESCRIPTOR* FindInterfaceDesc(const uint8_t *desc)
{
  const USB_COMMON_DESCRIPTOR *cur = (const USB_COMMON_DESCRIPTOR*)desc;
  
  while (cur->bLength != 0) {
    if (cur->bDescriptorType == USB_INTERFACE_DESCRIPTOR_TYPE) {
      return (const USB_INTERFACE_DESCRIPTOR*)cur;
    }
    cur = (const USB_COMMON_DESCRIPTOR*)(((const uint8_t*)cur) + cur->bLength);
  }
  return NULL;
}

static ErrorCode_t HidInitialize(const USB_HID_COMM_PARAM *param, uint32_t hidMemBase, uint32_t hidMemSize)
{
  USBD_HID_INIT_PARAM_T hidParam = {0};

  hidParam.max_reports = 1;
  hidParam.mem_base = hidMemBase;
  hidParam.mem_size = hidMemSize;
  hidParam.intf_desc = (uint8_t*)FindInterfaceDesc(param->configDesc);
  hidParam.HID_GetReport = GetReport;
  hidParam.HID_SetReport = SetReport;
  hidParam.report_data = param->reportData;

  return USBD_API->hid->init(mUsbHandle, &hidParam);
}

void USB_IRQHandler(void)
{
  USBD_API->hw->ISR(mUsbHandle);
}

void USBHidCommInitialize(const USB_HID_COMM_PARAM *param)
{
  uint32_t memBase, memSize;
  USBD_API_INIT_PARAM_T usbParam = {0};
  USB_CORE_DESCS_T usbDesc = {0};
  ErrorCode_t ret = LPC_OK;

  Chip_USB_Init();

  g_pUsbApi = (const USBD_API_T *) LPC_ROM_API->usbdApiBase;

  memBase = param->memBase;
  memSize = param->memSize;

  usbParam.usb_reg_base = LPC_USB0_BASE;
  usbParam.max_num_ep = 2; // 本当はEP0だけなので1を指定するが  Errata USB_ROM.1 のW.A.で+1
  usbParam.USB_Interface_Event = StallWorkAround;
  usbParam.mem_base = memBase;
  usbParam.mem_size = memSize;

  usbDesc.device_desc = (uint8_t*)param->deviceDesc;
  usbDesc.string_desc = (uint8_t*)param->stringDesc;
  usbDesc.high_speed_desc = (uint8_t*)param->configDesc;
  usbDesc.full_speed_desc = (uint8_t*)param->configDesc;
  usbDesc.device_qualifier = 0;

  ret = USBD_API->hw->Init(&mUsbHandle, &usbDesc, &usbParam);
  if (ret == LPC_OK) {
    
    // WORKAROUND for artf32219 ROM driver BUG:
    memBase = param->memBase + param->memSize - usbParam.mem_size;

    ret = HidInitialize(param, memBase, usbParam.mem_size);
    if (ret == LPC_OK) {

      mSendToHostHandler = param->sendToHostHandler;
      mRecvFromHostHandler = param->recvFromHostHandler;
      
      NVIC_EnableIRQ(USB0_IRQn);
      USBD_API->hw->Connect(mUsbHandle, 1); // 1: Connect
    }
  }
}
